/*
  nitftopnm - convert a National Imagery Transmission Format file to pnm

*/

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#include <unistd.h>
#include <string.h>

#include <nitf.h>

static int writeascii(FILE *ofile, char *desc, char *str, int len);
static int writeint(FILE *ofile, char *desc, int val);

static void usage(int ret)
{
  fprintf(ret?stderr:stdout, 
  "nitftopnm - convert a National Imagery Transmission Format file to pnm\n"
  "usage:   nitftoppm [options]\n"
  "options:\n"
  "   -o ofile, --output ofile  write output to ofile instead of stdout.\n"
  "   -i n, --image n  convert the nth image in the input file, instead of\n"
  "                    the first.\n"
  "   -h, --help       display this help\n"
  "   -v, --version    display program version\n");
  exit(ret);
}

int main (int argc, char **argv)
{
  struct nitf_header *hdr;
  int err;
  int c;
  FILE *ifile=stdin;
  FILE *ofile=stdout;
  int image=0;
  extern int optind;

  int longind;
  static struct option longopts[] =
    {
      {"help", 0, 0, 'h'},
      {"image", 1, 0, 'i'},
      {"output", 1, 0, 'o'},
      {"version", 1, 0, 'v'},
      {0, 0, 0, 0}
    };
  while ((c = getopt_long(argc, argv, "hio:v", longopts, &longind)) != EOF)

    {
      switch(c)
	{
	case 'h': usage(0);
	case 'i': image = atoi(optarg) - 1; break;
	case 'o': 
	  ofile = fopen(optarg, "w");
	  if (!ofile)
	    {
	      fprintf(stderr, "cannot open %s for writing\n", optarg);
	      exit(1);
	    }
	  break;
	case 'v': printf("nitftopnm version %s\n", VERSION); break;
	default: usage(1);
	}
    }

  if (optind <= argc)
    {
      ifile = fopen(argv[optind], "r");
      if (ifile == 0)
	{
	  fprintf(stderr, "cannot open %s\n", argv[optind]);
	  exit(1);
	}
    }

  err = readnitf(ifile, &hdr);

  writeppm(ofile, hdr, &hdr->ilink[image]);

  exit(err);
}


#define GETASCII(where, bytes)			\
  memmove(&where, next, bytes);			\
  next += bytes
#define GETINT(where, bytes)			\
  memmove(work, next, bytes);			\
  next += bytes;				\
  work[bytes]=0;				\
  where = atoi(work)

char *buf, *next, work[32];
int bufmax=0;
int bufcnt=0;

static int readnumber(FILE *ifile, int size, int *result);
static int readband(FILE *ifile, struct band_st **bandp);


/* read NITF header from file into a struct allocated from the heap.
  Return result code.  */
int readnitf(FILE *ifile, struct nitf_header **hdrp)
{
  int n;
  int i;
  int expected;
  int result;
  struct nitf_header *hdr;
  hdr = xmalloc(sizeof(struct nitf_header));
  buf = xrealloc(buf, bufmax = 8192);

  /* initial read:  everything up to the first conditional field */
  expected = 9+2+4+10+14+80+1+40+40+40+20+20+6;
  bufcnt = n = fread(buf, 1, expected, ifile);
  if (ferror(ifile) || feof(ifile)) return NITF_FAIL;
  if (n < expected) return NITF_FAIL;

  /* check whether fsdevt field is present */
  next = buf + bufcnt - 6;
  expected = 5+5+1+27+18+12+6;
  if (strncmp(next, "999998", 6) == 0) expected += 40;

  /* second read: through file header length */
  bufcnt += n = fread(buf+bufcnt, 1, expected, ifile);
  if (ferror(ifile) || feof(ifile)) return NITF_FAIL;
  if (n < expected) return NITF_FAIL;

  next = buf + bufcnt - 6;
  GETINT(hdr->hl, 6);
  if (hdr->hl > bufmax)
    buf = xrealloc(buf, bufmax = hdr->hl + 4096);

  /* read rest of header */
  expected = hdr->hl - bufcnt;
  bufcnt += n = fread(buf+bufcnt, 1, expected, ifile);
  if (ferror(ifile) || feof(ifile)) return NITF_FAIL;
  if (n < expected) return NITF_FAIL;

  /* parse the header */
  next = buf;
  GETASCII(hdr->fhdr, 9);	/* MIL-STD-2500A specifies that this
				   field must contain either
				   "NITF01.10" or "NITF02.00" */
  if (strncmp(hdr->fhdr, "NITF", 4) != 0)
    return NITF_BADFILE;

  GETASCII(hdr->clevel, 2);
  GETASCII(hdr->stype, 4);
  GETASCII(hdr->ostaid, 10);
  GETASCII(hdr->fdt, 14);	/* MIL-STD-2500A specifies that the
				   date has the format 23163547ZFEB99
				   (which starts with day of month,
				   hours, minutes, and seconds).  This
				   allows only two digits for the year
				   and so is not Y2K compliant. Presumably
				   there is a later version of the
				   standard that changes this, but I
				   have not seen it. */
  GETASCII(hdr->ftitle, 80);
  GETASCII(hdr->fsclas, 1);
  GETASCII(hdr->fscode, 40);
  GETASCII(hdr->fsctlh, 40);
  GETASCII(hdr->fsrel, 40);
  GETASCII(hdr->fscaut, 20);
  GETASCII(hdr->fsctln, 20);
  GETASCII(hdr->fsdwng, 6);
  /* conditional field fsdevt, file downgrading event */
  if (strncmp(hdr->fsdwng, "999998", 6) == 0)
    {
      GETASCII(hdr->fsdevt, 40);
    }
  GETINT(hdr->fscop, 5);
  GETINT(hdr->fscpys, 5);
  GETASCII(hdr->encryp, 1);
  GETASCII(hdr->oname, 27);
  GETASCII(hdr->ophone, 18);
  GETINT(hdr->fl, 12);
  GETINT(hdr->hl, 6);

  /* image lengths */
  GETINT(hdr->numi, 3);
  hdr->ilink = xmalloc(hdr->numi*sizeof(struct image_link));
  for (i = 0; i < hdr->numi; i++)
    {
      GETINT(hdr->ilink[i].lish, 6);
      GETINT(hdr->ilink[i].li, 10);
    }

  /* symbol lengths */
  GETINT(hdr->nums, 3);
  if (hdr->nums)
    {
      hdr->slink = xmalloc(hdr->nums*sizeof(struct symbol_link));
      for (i = 0; i < hdr->nums; i++)
	{
	  GETINT(hdr->slink[i].lssh, 6);
	  GETINT(hdr->slink[i].ls, 10);
	}
    }

  /* label lengths */
  GETINT(hdr->numl, 3);
  if (hdr->numl)
    {
      hdr->llink = xmalloc(hdr->nums*sizeof(struct label_link));
      for (i = 0; i < hdr->numl; i++)
	{
	  GETINT(hdr->llink[i].llsh, 6);
	  GETINT(hdr->llink[i].ll, 10);
	}
    }

  /* text lengths */
  GETINT(hdr->numt, 3);
  if (hdr->numt)
    {
      hdr->tlink = xmalloc(hdr->numt*sizeof(struct text_link));
      for (i = 0; i < hdr->numt; i++)
	{
	  GETINT(hdr->tlink[i].ltsh, 6);
	  GETINT(hdr->tlink[i].lt, 10);
	}
    }

  /* des lengths */
  GETINT(hdr->numd, 3);
  if (hdr->numd)
    {
      hdr->dlink = xmalloc(hdr->numd*sizeof(struct des_link));
      for (i = 0; i < hdr->numd; i++)
	{
	  GETINT(hdr->dlink[i].ldsh, 6);
	  GETINT(hdr->dlink[i].ld, 10);
	}
    }

  /* res lengths */
  GETINT(hdr->numr, 3);
  if (hdr->numr)
    {
      hdr->rlink = xmalloc(hdr->numr*sizeof(struct res_link));
      for (i = 0; i < hdr->numr; i++)
	{
	  GETINT(hdr->rlink[i].lrsh, 6);
	  GETINT(hdr->rlink[i].lr, 10);
	}
    }

  GETINT(hdr->udhdl, 5);
  if (hdr->udhdl)
    {
      GETINT(hdr->udhofl, 3);
      GETASCII(hdr->udhd, hdr->udhdl);
    }
  GETINT(hdr->xhdl, 5);
  GETASCII(hdr->xhd, hdr->xhdl);

  /* images */
  for (i = 0; i < hdr->numi; i++)
    {
      result = readimage(ifile, &hdr->ilink[i]);
      if (result) return result;
    }

  /* read symbols FIXME */

  /* read labels FIXME */

  /* read text FIXME */

  /* read des FIXME */

  /* read res FIXME */

  *hdrp = hdr;
  return NITF_OK;
}

/* read one image subheader and image */
int readimage(FILE *ifile, struct image_link *ilink)
{
  int expected;
  int n;
  int i;

  struct image_subheader *isubh = xmalloc(ilink->lish);

  /* fields fhdr through fsdwng */
  expected = ilink->lish;
  if (expected > bufmax)
    buf = xrealloc(buf, bufmax = expected + 4096);
  bufcnt = n = fread(buf, 1, expected, ifile);
  if (ferror(ifile) || feof(ifile)) return NITF_FAIL;
  if (n < expected) return NITF_FAIL;
  next = buf;

  GETASCII(isubh->im, 2);
  if (strncmp(isubh->im, "IM", 2) != 0) return NITF_BADFILE;
  GETASCII(isubh->iid, 10);
  GETASCII(isubh->idatim, 14);
  GETASCII(isubh->tgtid, 17);
  GETASCII(isubh->ititle, 80);
  GETASCII(isubh->isclas, 1);
  GETASCII(isubh->iscode, 40);
  GETASCII(isubh->isctlh, 40);
  GETASCII(isubh->isrel, 40);
  GETASCII(isubh->iscaut, 20);
  GETASCII(isubh->isctln, 20);
  GETASCII(isubh->isdwng, 6);
  /* conditional field fsdevt, file downgrading event */
  if (strncmp(isubh->isdwng, "999998", 6) == 0)
    {
      GETASCII(isubh->isdevt, 40);
    }
  GETASCII(isubh->encryp, 1);
  GETASCII(isubh->isorce, 42);
  GETINT(isubh->nrows, 8);
  GETINT(isubh->ncols, 8);
  GETASCII(isubh->pvtype, 3);
  GETASCII(isubh->irep, 8);
  GETASCII(isubh->icat, 8);
  GETINT(isubh->abpp, 2);
  GETASCII(isubh->pjust, 1);
  GETASCII(isubh->icords, 1);
  /* conditional field igeolo, image geographic location */
  if (isubh->icords[0] != 'N')
    {
      GETASCII(isubh->igeolo, 60);
    }
  GETINT(isubh->nicom, 1);
  /* conditional fields icom1 ... icomnn (comments) */
  if (isubh->nicom)
    {
      isubh->icom = xmalloc(isubh->nicom*80);
      for (i = 0; i < isubh->nicom; i++)
	{
	  GETASCII(isubh->icom+i*80, 80);
	}
    }
  GETASCII(isubh->ic, 2);
  if (strncmp(isubh->ic, "NC", 2) && strncmp(isubh->ic, "NM", 2))
    {
      GETASCII(isubh->comrat, 4);
    }
  GETINT(isubh->nbands, 1);

  if (isubh->nbands)
    {
      isubh->band = xmalloc(isubh->nbands*sizeof(struct band_st));

      /* fields for one band: irepband through lutd */
      for (i = 0; i < isubh->nbands; i++)
	{
	  GETASCII(isubh->band[i].irepband, 2);
	  GETASCII(isubh->band[i].isubcat, 6);
	  GETASCII(isubh->band[i].ifc, 1);
	  GETASCII(isubh->band[i].imflt, 3);
	  GETINT(isubh->band[i].nluts, 1);
	  if (isubh->band[i].nluts)
	    {
	      GETINT(isubh->band[i].nelut, 5);
	      isubh->band[i].lutd = 
		xmalloc(isubh->band[i].nluts*isubh->band[i].nelut);
	      GETASCII(isubh->band[i].lutd, 
		       isubh->band[i].nluts*isubh->band[i].nelut);
	    }
	}
    }
  GETINT(isubh->isync, 1);
  GETASCII(isubh->imode, 1);
  GETINT(isubh->nbpr, 4);
  GETINT(isubh->nbpc, 4);
  GETINT(isubh->nppbh, 4);
  GETINT(isubh->nppbv, 4);
  GETINT(isubh->nbpp, 2);
  GETINT(isubh->idlvl, 3);
  GETINT(isubh->ialvl, 3);
  GETINT(isubh->ilocr, 5);
  GETINT(isubh->ilocc, 5);
  GETASCII(isubh->imag, 4);
  GETINT(isubh->udidl, 5);
  if (isubh->udidl)
    {
      GETINT(isubh->udofl, 3);
      isubh->udid = xmalloc(isubh->udidl);
      GETASCII(isubh->udid, isubh->udidl);
    }
  GETINT(isubh->ixshdl, 5);
  if (isubh->ixshdl)
    {
      GETINT(isubh->ixsofl, 3);
      isubh->ixshd = xmalloc(isubh->ixshdl);
      GETASCII(isubh->ixshd, isubh->ixshdl);
    }

  ilink->ish = isubh;

  /* read an image */
  expected = ilink->li;
  ilink->im = xmalloc(expected);
  n = fread(ilink->im, 1, expected, ifile);
  if (ferror(ifile) || feof(ifile)) return NITF_FAIL;
  if (n < expected) return NITF_FAIL;

  return NITF_OK;
}

/*
  writeppm - write a Portable Pixmap using data from a populated NITF struct

*/

int writeppm(FILE *ofile, struct nitf_header *hdr, struct image_link *ilink)
{
  char obuf[80];
  int n;
  int expected;
  int nrows, ncols;

  struct image_subheader *isubh = ilink->ish;

  nrows = isubh->nrows;
  ncols = isubh->ncols;

  if (strncmp(isubh->ic, "NC", 2))
    fail("sorry, compressed images are not supported");

  if ((strncmp(isubh->irep,"RGB",3) == 0 && 
       isubh->nbands == 3  &&
       isubh->imode[0] == 'P') ||
      (strncmp(isubh->irep,"RGB/LUT",7) == 0 && 
       isubh->nbands == 1 &&
       isubh->band[0].nluts == 3 &&
       isubh->imode[0] == 'P'))
    fprintf(ofile, "P6 %d %d\n", ncols, nrows);
  else if (strncmp(isubh->irep,"MONO",4) == 0 && 
	   isubh->nbands == 1 &&
	   isubh->imode[0] == 'P')
    fprintf(ofile, "P2 %d %d\n", ncols, nrows);
  else 
    {
      fprintf(stderr, 
	  "Sorry, this image format (%.8s with %d bands and mode %.1s) "
	      "is not supported.\n"
	  "Currently only these image representations are supported:\n"
	  "RGB, RGB/LUT, and MONO", 
	      isubh->irep, isubh->nbands, isubh->imode);
      exit(1);
    }

  fprintf(ofile, "# Converted from NITF\n");

  writeascii(ofile, "fhdr (file type and version)", hdr->fhdr, 9);
  writeascii(ofile, "stype", hdr->stype, 4);
  writeascii(ofile, "ostaid", hdr->ostaid, 10);
  writeascii(ofile, "fdt", hdr->fdt, 14);
  writeascii(ofile, "ftitle", hdr->ftitle, 80);
  writeascii(ofile, "fsclas (file security classification)", hdr->fsclas, 1);
  writeascii(ofile, "fscode", hdr->fscode, 40);
  writeascii(ofile, "fsctlh", hdr->fsctlh, 40);
  writeascii(ofile, "fsrel", hdr->fsrel, 40);
  writeascii(ofile, "fscaut", hdr->fscaut, 20);
  writeascii(ofile, "fsctln", hdr->fsctln, 20);
  writeascii(ofile, "fsdwng", hdr->fsdwng, 6);
  if (strncmp(hdr->fsdwng, "999998", 6) == 0)
    {
      writeascii(ofile, "fsdevt", hdr->fsdevt, 40);
    }
  writeascii(ofile, "oname", hdr->oname, 27);
  writeascii(ofile, "ophone", hdr->ophone, 18);
  writeint(ofile, "numi (number of images)", hdr->numi);
  writeascii(ofile, "iid", isubh->iid, 10);
  writeascii(ofile, "idatim", isubh->idatim, 14);
  writeascii(ofile, "tgtid", isubh->tgtid, 17);
  writeascii(ofile, "ititle", isubh->ititle, 80);
  writeascii(ofile, "isclas (image security classification)", 
	     isubh->isclas, 1);
  writeascii(ofile, "iscode", isubh->iscode, 40);
  writeascii(ofile, "isctlh", isubh->isctlh, 40);
  writeascii(ofile, "isrel", isubh->isrel, 40);
  writeascii(ofile, "iscaut", isubh->iscaut, 20);
  writeascii(ofile, "isctln", isubh->isctln, 20);
  writeascii(ofile, "isdwng", isubh->isdwng, 6);
  /* conditional field fsdevt, file downgrading event */
  if (strncmp(isubh->isdwng, "999998", 6) == 0)
    {
      writeascii(ofile, "isdevt", isubh->isdevt, 40);
    }
  writeascii(ofile, "isorce (image source)", isubh->isorce, 42);
  writeascii(ofile, "irep", isubh->irep, 8);
  writeascii(ofile, "icat", isubh->icat, 8);
  writeint(ofile, "abpp (actual bits per pixel per band)", isubh->abpp);
  writeascii(ofile, "icords (image coordinate system)", isubh->icords, 1);
  if (isubh->icords[0] != 'N')
    {
      writeascii(ofile, "igeolo", isubh->igeolo, 60);
      writeascii(ofile, "(NW corner)", isubh->igeolo, 15);
      writeascii(ofile, "(NE corner)", isubh->igeolo+15, 15);
      writeascii(ofile, "(SE corner)", isubh->igeolo+30, 15);
      writeascii(ofile, "(SW corner)", isubh->igeolo+45, 15);
    }
  if (isubh->nicom)
    {
      int i;
      for (i = 0; i < isubh->nicom; i++)
	writeascii(ofile, "", isubh->icom+i*80, 80);
    }
  writeascii(ofile, "ic (image compression)", isubh->ic, 2);
  if (strncmp(isubh->ic, "NC", 2) && strncmp(isubh->ic, "NM", 2))
    {
      writeascii(ofile, "comrat", isubh->comrat, 4);
    }
  writeint(ofile, "nbands", isubh->nbands);
  writeint(ofile, "isync", isubh->isync);
  writeascii(ofile, "imode", isubh->imode, 1);
  
  fprintf(ofile, "%d\n", 255);

				/* write the image data */
  if (strncmp(isubh->irep,"RGB",3) == 0 &&
      isubh->nbands == 3 &&
      isubh->imode[0] == 'P')
    {
      expected = ncols*nrows*isubh->nbands;
      n = fwrite(ilink->im, 1, expected, ofile);
      if (n != expected) return NITF_FAIL;
    }
  else if (strncmp(isubh->irep,"RGB/LUT",7) == 0 &&
      isubh->nbands == 1 &&
      isubh->band[0].nluts == 3 &&
      isubh->imode[0] == 'P')
    {
      char *row = xmalloc(ncols*3);
      char *lutd = isubh->band[0].lutd;
      int nelut = isubh->band[0].nelut;
      int r, c, i, k, code;
      for (k = r = 0; r < nrows; r++)
	{
	  for (i = c = 0; c < ncols; c++)
	    {
	      code = ilink->im[k++];
	      row[i++] = lutd[0*nelut + code];
	      row[i++] = lutd[1*nelut + code];
	      row[i++] = lutd[2*nelut + code];
	    }
	  expected = ncols*3;
	  n = fwrite(ilink->im, 1, expected, ofile);
	  if (n != expected) return NITF_FAIL;
	}
      free(row);
    }
  else if (strncmp(isubh->irep,"MONO",4) == 0 &&
	   isubh->nbands == 1 &&
	   isubh->imode[0] == 'P')
    {
      expected = ncols*nrows;
      n = fwrite(ilink->im, 1, expected, ofile);
      if (n != expected) return NITF_FAIL;
    }
  else 
    fail("sorry, this image format is not supported\n"); /* doesn't return */


  return NITF_OK;
}

/* If the len characters starting at str contain nonblanks, write them
   to ofile as a comment.  */
static int writeascii(FILE *ofile, char *desc, char *str, int len)
{
  char *s;
  int expected;
  static char *obuf = 0;
  static int obufmax = 0;
  int n;

  if (obufmax < len + strlen(desc) + 10)
    obuf = xrealloc(obuf, obufmax = len + strlen(desc) + 20);

  for (s = str + len; --s >= str && *s == ' '; ) {}
  expected = s - str + 1;

  if (expected > 0)
    {
      n = fprintf(ofile, "# %s: %.*s\n", desc, expected, str);
      if (n < 5) return NITF_FAIL;
    }
  return NITF_OK;
}

/* write val to ofile as a comment */
static int writeint(FILE *ofile, char *desc, int val)
{
  char *s;
  int expected;
  static char *obuf = 0;
  static int obufmax = 0;
  int n;

  if (obufmax < strlen(desc) + 10)
    obuf = xrealloc(obuf, obufmax = strlen(desc) + 20);

  n = fprintf(ofile, "# %s: %d\n", desc, val);
  if (n < 5) return NITF_FAIL;

  return NITF_OK;
}


/*
   utilities 
*/

void *xrealloc(void *ptr, unsigned n)
{
  void *p = realloc(ptr, n);
  if (!p)
    {
      perror("nitf");
      exit(1);
    }
  return p;
}

void *xmalloc(unsigned n)
{
  return xrealloc(0, n);
}

void fail(char *s)
{
  fprintf(stderr, "%s\n", s);
  exit(1);
}
